import functools
import re
import random
import sys
from collections import Counter
from more_itertools import grouper

PERFECT = "\N{LARGE GREEN SQUARE}"
GOOD = "\N{LARGE YELLOW SQUARE}"
MISS = "\N{WHITE LARGE SQUARE}"
WORDS = [s for a in "+-<>,.[]" for b in "+-<>,.[]" for c in "+-<>,.[]" for d in "+-<>,.[]" for e in "+-<>,.[]" if all(x not in (s:=ѕ) for ѕ in [a+b+c+d+e] for y in ("+-", "<>") for x in (y, y[::-1]))]

@functools.cache
def match(candidate, target):
    counts = Counter(target)
    results = []
    for c, t in zip(candidate, target):
        if c == t:
            results.append(PERFECT)
            counts[c] -= 1
        else:
            results.append(None)
    for i, c in enumerate(candidate):
        if results[i] is None:
            if counts[c]:
                results[i] = GOOD
                counts[c] -= 1
            else:
                results[i] = MISS
    return "".join(results)

def result_to_byte(result):
    n = 0
    for char in result:
        n *= 3
        n += 0 if char == MISS else 1 if char == GOOD else 2
    return n

def byte_to_result(byte):
    result = ""
    for _ in range(5):
        result += [MISS, GOOD, PERFECT][byte%3]
        byte //= 3
    return result[::-1]

def write_code(code, file):
    while code != (next_code := re.sub(r"[^+\-<>,.[\]]|\+-|-\+|<>|><", "", code)):
        code = next_code
    for chunk in grouper(code, 5, "+"):
        table = []
        for word in WORDS:
            table.append(result_to_byte(match(word, "".join(chunk))))
        file.write(bytes(table))

@functools.cache
def solve(chunk):
    pool = list(enumerate(WORDS))
    while len(pool) > 1:
        i, guess = random.choice(pool)
        answer = byte_to_result(chunk[i])
        pool = [(i, word) for i, word in pool if match(guess, word) == answer]
    return pool[0][1]

def read_code(file):
    code = ""
    while chunk := file.read(len(WORDS)):
        code += solve(chunk)
    return code

def interpret_bf(code):
    while code != (next_code := re.sub(r"[^+\-<>,.[\]]|\+-|-\+|<>|><", "", code)):
        code = next_code
    to_execute = list(code)
    tape = [0]*30_000
    head = 0
    for i, inst in enumerate(to_execute):
        match inst[0]:
            case "+":
                tape[head] += 1
            case "-":
                tape[head] -= 1
            case ">":
                head += 1
            case "<":
                head -= 1
            case ",":
                if b := sys.stdin.buffer.read(1):
                    tape[head] = b[0]
            case ".":
                sys.stdout.buffer.write(bytes([tape[head]]))
                if tape[head] == 10:
                    sys.stdout.buffer.flush()
            case "[":
                level = 0
                for j, c in enumerate(to_execute[i+1:], start=i+1):
                    to_execute[j] = c.translate(str.maketrans("+-<>,.[]", "ABCDEFGH") if not tape[head] else str.maketrans("ABCDEFGH", "+-<>,.[]"))
                    match c[0]:
                        case "]" | "H" if level == 0:
                            if tape[head]:
                                to_execute[j] = "]" + str(i)
                            break
                        case "]" | "H":
                            level -= 1
                        case "[":
                            level += 1
            case "]":
                j = int(inst[1:])
                for k, c in enumerate(to_execute[j:i+1], start=i+1):
                    to_execute.insert(k, c)

if __name__ == "__main__":
    match sys.argv:
        case (_, "-c", infile, outfile):
            with open(infile) as f, open(outfile, "wb") as o:
                write_code(f.read(), o)
        case (_, "-r", file):
            with open(file, "rb") as f:
                code = read_code(f)
            interpret_bf(code)
        case _:
            print("invalid arguments", file=sys.stderr)    
